import contextlib
import datetime
import logging
import os
import re
import shutil
from collections.abc import Generator, Iterator
from pathlib import Path
from typing import Any, Optional

from rich.progress import Progress, TaskID, TimeElapsedColumn

OptionalPath = Optional[Path]
OptionalIntList = Optional[list[int]]
OptionalStrList = Optional[list[str]]
OptionalPathList = Optional[list[Path]]


def create_path(
    *args: str,
    absolute: bool = True,
    dt_fmt: str = "%Y_%m_%d_%H_%M_%S",
    dt: datetime.datetime | None = None,
    **kwargs: Any,
) -> bytes | str:
    """
    Construct a path.  You can access dt or datetime as objects.

    :param args: path components passed to :py:func:`os.path.join`
    :param absolute: make path absolute
    :param dt: datetime object
    :param dt_fmt: date format
    :param kwargs: variables passed to :py:func:`str.format`

    :type absolute: bool
    :type dt: :py:class:`datetime.datetime`
    :type dt_fmt: str
    """
    if not dt:
        dt = datetime.datetime.now()

    dt_str = dt.strftime(dt_fmt)
    _kwargs = {"dt": dt, "datetime": dt, "date": dt_str}
    _kwargs.update(**kwargs)

    path = os.path.expanduser(os.path.join(*args).format(**_kwargs))

    if absolute:
        return os.path.abspath(path)

    return path


def backup(path: str | Path, copy: bool = False) -> Path | None:
    """
    Create backup file name.

    :param path: path of file
    :param copy: perform copy file
    """
    old_path = os.path.abspath(os.path.expanduser(path))

    # nothing to back up
    if not os.path.exists(old_path):
        return None

    if not os.path.isfile(old_path):
        raise RuntimeError(f"can only backup files: {old_path}")

    # extract file info
    dname, bname = os.path.dirname(old_path), os.path.basename(old_path)
    root, dirs, files = next(os.walk(dname))

    # look for existing backups
    regex = re.compile(rf"^{bname}(\.(\d+))?$")
    found = []
    for f in files:
        match = regex.match(f)
        if match:
            try:
                found.append(int(match.group(2)))
            except TypeError:
                found.append(0)

    # should have at least found 1 if the file exists
    if not found:
        raise RuntimeError(f"can not backup file: {old_path}")

    # create backup name
    new_path = os.path.join(dname, f"{bname}.{max(found) + 1}")

    # validate old_path
    if os.path.exists(new_path) or new_path == old_path:
        raise RuntimeError(f"can not backup file: {old_path}")

    # copy the file
    if copy:
        logging.debug("backup (old): %s", old_path)
        logging.debug("backup (new): %s", new_path)
        shutil.copy(old_path, new_path)

    return Path(new_path)


def find_files(
    top: str, regex: str | re.Pattern[str] = r".*", r: bool = False, ignorecase: bool = True, **kwargs: Any
) -> Iterator[str]:
    """
    Search for files that match pattern.

    :param ignorecase: ignore case in regular expression
    :param top: top level directory
    :param regex: pattern to match
    :param r: recursive search
    """
    regex = (
        regex
        if isinstance(regex, re.Pattern)
        else (re.compile(regex, re.IGNORECASE) if ignorecase else re.compile(regex))
    )

    if r:
        for root, _dirs, files in os.walk(top, **kwargs):
            for f in files:
                match = regex.match(f)
                if match:
                    yield os.path.join(root, f)
    else:
        root, dirs, files = next(os.walk(top, **kwargs))
        for f in files:
            match = regex.match(f)
            if match:
                yield os.path.join(root, f)


@contextlib.contextmanager
def progress_bar(description: str, total: int) -> Generator[tuple[Progress, TaskID]]:
    try:
        with Progress(
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            transient=True,
        ) as progress:
            task = progress.add_task(description=description, total=total)
            yield progress, task
    finally:
        pass



from __future__ import annotations

import asyncio
import contextlib
import json
import os
import re
from collections.abc import Iterable
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Any, cast

import aiohttp
import typer
from bs4 import Tag

from ffxiahbot.common import OptionalPath, progress_bar
from ffxiahbot.logutils import logger
from ffxiahbot.scrubbing.enums import ServerID
from ffxiahbot.scrubbing.scrubber import Scrubber

REGEX_CATEGORY: re.Pattern = re.compile(r"/browse/(\d+)/?.*")
REGEX_ITEM: re.Pattern = re.compile(r"/item/(\d+)")
REGEX_NAME: re.Pattern = re.compile(r"(.*?)\s*-?\s*(FFXIAH)?\.(com)?")


FailedType = dict[int, Exception]
ResultType = dict[int, dict[str, Any]]
ItemIDsType = list[int] | set[int] | tuple[int, ...]
CatURLsType = list[str] | set[str] | tuple[str, ...]


@dataclass()
class FFXIAHScrubber(Scrubber):
    """
    Get item data from the ffxiah.com website.
    """

    #: The FFXI server
    server: ServerID = ServerID.BAHAMUT
    #: The cache directory
    cache: OptionalPath = None

    async def scrub(
        self,
        cat_urls: CatURLsType | None = None,
        item_ids: ItemIDsType | None = None,
    ) -> tuple[ResultType, FailedType]:
        """
        Get item metadata main function.

        The item ids can be loaded from category urls or simply passed as a list.  The urls
        can be generated automatically, in which case all possible items will be downloaded.

        Args:
            cat_urls: A preset list of category URLs.
            item_ids: A preset list of item IDs.

        Returns:
            failed: failed item ids
            data: item data
        """
        async with aiohttp.ClientSession() as session:
            if not item_ids:
                if not cat_urls:
                    cat_urls = await self._get_category_urls(session)

                logger.debug("# urls = %d", len(cat_urls))
                item_ids = await self._get_item_ids(session, cat_urls)
            else:
                logger.debug("using passed ids")
                item_ids = sorted(set(item_ids))

                if cat_urls:
                    logger.warning("passed urls ignored")

            results, failed = await self._get_item_data(session, item_ids)
            logger.debug("expect count = %d", len(item_ids))
            logger.debug("passed count = %d", len(results))
            logger.debug("failed count = %d", len(failed))

            return results, failed

    # step 1
    async def _get_category_urls(self, session: aiohttp.ClientSession) -> list[str]:
        """
        Parse http://www.ffxiah.com/browse to get URLs of the form http://www.ffxiah.com/{CategoryNumber}.
        """
        logger.debug("getting category urls")

        # the browse section of FFXIAH has a list of urls with category numbers
        logger.debug("open %s", path := "http://www.ffxiah.com/browse")
        soup = await self.soup(session, path)

        def _() -> Iterable[str]:
            for tag in soup.find_all("a"):
                if tag.has_attr("href"):
                    href = tag.get("href")
                    match = REGEX_CATEGORY.match(href)
                    if match:
                        try:
                            category = int(match.group(1))
                            if category < 240:
                                yield f"http://www.ffxiah.com{href}"
                                logger.debug("category %s", href)
                            else:
                                logger.debug("skipping %s", href)

                        except (ValueError, IndexError):
                            logger.exception("failed to extract category")
                    else:
                        logger.debug("ignoring %s", href)

        return sorted(_(), key=lambda x: list(map(float, re.findall(r"\d+", x))))

    # step 2
    async def _get_item_ids(self, session: aiohttp.ClientSession, urls: CatURLsType) -> list[int]:
        """
        Scrub urls of the form http://www.ffxiah.com/{CategoryNumber} for itemids.
        """
        tasks = []
        logger.info("getting itemids")
        for i, cat_url in enumerate(urls):
            logger.info("category %02d/%02d : %s", i + 1, len(urls), cat_url)
            tasks.append(asyncio.create_task(self._get_itemids_for_category_url(session, cat_url)))

        return sorted(set(chain(*await asyncio.gather(*tasks))))

    # step 2.1
    async def _get_itemids_for_category_url(self, session: aiohttp.ClientSession, url: str) -> Iterable[int]:
        """
        Scrub url of the form http://www.ffxiah.com/{CategoryNumber} for itemids.
        """
        # create tag soup
        soup = await self.soup(session, url)

        def _() -> Iterable[int]:
            # look for table class
            table = soup.find("table", class_="stdlist")
            if not table:
                logger.error("failed to parse <table>")
                return

            # look for table body
            tbody = table.find("tbody")
            if not tbody:
                logger.error("failed to parse <tbody>")
                return

            # look for table rows
            trs = cast(Tag, tbody).find_all("tr")
            if not trs:
                logger.error("failed to parse <tr>")
                return

            # loop table rows
            count = 0
            for j, row in enumerate(trs):
                # look for 'a' tag
                a = row.find("a")

                if a is not None:
                    # look for href attr
                    href = a.get("href")

                    if href is not None:
                        # make sure href matches /item/{number}
                        if (match := REGEX_ITEM.match(href)) is None:
                            logger.error("failed to extract itemid!\n\n\trow %d of %s\n\n%s\n\n", j, url, row)
                        else:
                            item = int(match.group(1))
                            count += 1
                            yield item
                    else:
                        logger.error("failed to extract href!\n\n\trow %d of %s\n\n%s\n\n", j, url, row)
                else:
                    logger.error("failed to extract 'a' tag!\n\n\trow %d of %s\n\n%s\n\n", j, url, row)

            # make sure we found something
            if not count:
                logger.error("could not find any itemids!")

        return set(_())

    # step 3
    async def _get_item_data(self, session: aiohttp.ClientSession, itemids: list[int]) -> tuple[ResultType, FailedType]:
        """
        Get metadata for many items.
        """
        logger.info("getting data")
        tasks = {
            itemid: asyncio.create_task(self._get_item_data_for_itemid(session, itemid, index=i, total=len(itemids)))
            for i, itemid in enumerate(itemids)
        }

        results, failed = {}, {}
        with progress_bar("[red]Fetching Items...", total=len(itemids)) as (progress, progress_task):
            for i, (itemid, task) in enumerate(tasks.items()):
                progress.update(progress_task, advance=1)
                # noinspection PyBroadException
                try:
                    results[itemid] = await task
                    logger.info("item %06d/%06d : %06d", i, len(itemids), itemid)
                except Exception as e:
                    failed[itemid] = e
                    logger.exception("item %06d/%06d : %06d", i, len(itemids), itemid)
                    if len(failed) > int(os.environ.get("FFXIAHBOT_SCRUB_MAX_FAILURES", 10)):
                        logger.error("too many failures!")
                        raise typer.Exit(-1) from None
        return results, failed

    def _cache_sub(self, itemid: int) -> Path:
        """
        Create cache subdirectory.
        """
        if self.cache is None:
            raise RuntimeError("cache is not set!")

        return cast(Path, self.cache).joinpath(f"{self.server.value}", f"{itemid // 1000}", f"{itemid}")

    # step 3.1
    async def _get_item_data_for_itemid(
        self, session: aiohttp.ClientSession, itemid: int, index: int = 0, total: int = 0
    ) -> dict[str, Any]:
        """
        Get metadata for single item.
        """
        percent = float(index) / float(total) * 100.0 if total > 0 else 0.0

        data: dict[str, Any] = {"name": None, "itemid": itemid}
        url = f"http://www.ffxiah.com/item/{itemid}"

        # load cache
        if self.cache is not None:
            cache_stub = self._cache_sub(itemid)
            if (cache_path := cache_stub.with_suffix(".json")).exists():
                logger.debug("load server=%s (%06d/%06d,%6.2f) %s", self.server, index, total, percent, url)
                return cast(dict[str, Any], json.loads(cache_path.read_text()))

        # create tag soup
        logger.debug("open server=%s (%06d/%06d,%6.2f) %s", self.server, index, total, percent, url)
        soup = await self.soup(session, url, data={"sid": self.server.value})

        # extract name
        if (
            soup.title is not None
            and soup.title.text is not None
            and (name_match := REGEX_NAME.match(soup.title.text)) is not None
        ):
            data.update(name=name_match.group(1))
        else:
            data.update(name=None)

        # extract numbers
        for tag in soup.find_all("span", "number-format"):
            try:
                key = tag.parent.find_previous_sibling().text.lower()
                data[key] = int(float(tag.text))
            except (AttributeError, ValueError):
                pass

        # extract rate
        for tag in soup.find_all("span", "sales-rate"):
            with contextlib.suppress(AttributeError, ValueError):
                data["rate"] = float(tag.text)

        # fix key
        data = self._fix_stack_price_key(data)

        # save cache
        if self.cache is not None:
            cache_stub = self._cache_sub(itemid)
            cache_stub.parent.mkdir(parents=True, exist_ok=True)
            cache_stub.with_suffix(".json").write_text(json.dumps(data, indent=2, sort_keys=True))

        return data

    # step 3.1.2
    @staticmethod
    def _fix_stack_price_key(data: dict[str, Any]) -> dict[str, Any]:
        """
        Fix dictionary key.
        """
        new_key = r"stack price"

        for key in list(data.keys()):
            if "stack" in key.lower():
                data[new_key] = data[key]

        return data


def augment_item_info(item_info: dict[str, Any], **kwargs: Any) -> dict:
    """
    Extract item data from scrubbed info.
    """
    # singles
    try:
        price_single, sell_single = item_info["median"], True

        # do not sell items without a price
        if price_single <= 0:
            price_single, sell_single = 0, False

    except KeyError:
        price_single, sell_single = 0, False

    # stacks
    try:
        price_stacks, sell_stacks = item_info["stack price"], True

        # do not sell items without a price
        if price_stacks <= 0:
            price_stacks, sell_stacks = 0, False

    except KeyError:
        price_stacks, sell_stacks = 0, False

    # the name doesn't really matter
    try:
        name = item_info["name"]
    except KeyError:
        name = None

    result = {
        "name": name,
        "price_single": price_single,
        "stock_single": 5,
        "sell_single": sell_single,
        "buy_single": True,
        "rate_single": 1.0,
        "price_stacks": price_stacks,
        "stock_stacks": 5,
        "sell_stacks": sell_stacks,
        "buy_stacks": True,
        "rate_stacks": 1.0,
    }
    result.update(**kwargs)

    return result



import asyncio
import logging
from dataclasses import dataclass
from typing import Any

import aiohttp
import bs4
from bs4 import BeautifulSoup

TIMEOUT: int = 1024


@dataclass()
class Scrubber:
    # noinspection PyBroadException
    @staticmethod
    async def soup(session: aiohttp.ClientSession, url: str, **kwargs: Any) -> BeautifulSoup:
        """
        Open URL and create tag soup.

        Parameters:
            session: client session
            url: website string to open
        """
        max_tries = 10
        for i in range(max_tries):
            # noinspection PyPep8
            try:
                async with session.get(url, **kwargs) as response:
                    text = await response.text()
                break
            except Exception:
                logging.exception("urlopen failed (attempt %d)", i + 1)
                if i == max_tries - 1:
                    logging.exception("the maximum urlopen attempts have been reached")
                    raise
                await asyncio.sleep(1)

        try:
            s = BeautifulSoup(text, features="html5lib")
        except bs4.FeatureNotFound:
            s = BeautifulSoup(text, features="html.parser")

        return s



import re
from subprocess import run

from mkdocs_macros.plugin import MacrosPlugin

# fmt: off
# https://stackoverflow.com/questions/14693701
ANSI_ESCAPE_8BIT = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
    """,
    re.VERBOSE,
)
# fmt: on


def define_env(env: MacrosPlugin) -> None:
    @env.macro
    def get_help_message(*args: str):
        command = ["uv", "run", "python", "-m", *args, "--help"]
        message = run(command, capture_output=True).stdout.decode("utf-8")
        message = ANSI_ESCAPE_8BIT.sub("", message)
        return f"```bash\n➜  {' '.join(command)}\n{message}\n```"



import aiohttp
import pytest

from ffxiahbot.scrubbing.ffxiah import FFXIAHScrubber, augment_item_info


@pytest.mark.asyncio
async def test_get_category_urls():
    async with aiohttp.ClientSession() as session:
        await FFXIAHScrubber()._get_category_urls(session)


@pytest.mark.asyncio
async def test_get_itemids_for_category_url():
    url = r"http://www.ffxiah.com/browse/49/ninja-tools"
    async with aiohttp.ClientSession() as session:
        await FFXIAHScrubber()._get_itemids_for_category_url(session, url)


@pytest.mark.asyncio
async def test_get_itemids():
    urls = [
        r"http://www.ffxiah.com/browse/49/ninja-tools",
        r"http://www.ffxiah.com/browse/56/breads-rice",
    ]
    async with aiohttp.ClientSession() as session:
        await FFXIAHScrubber()._get_item_ids(session, urls)


@pytest.mark.asyncio
async def test_get_item_data_for_itemid():
    async with aiohttp.ClientSession() as session:
        await FFXIAHScrubber()._get_item_data_for_itemid(session, 4096)


@pytest.mark.asyncio
async def test_get_item_data():
    async with aiohttp.ClientSession() as session:
        await FFXIAHScrubber()._get_item_data(session, list(range(1, 9)))


@pytest.mark.asyncio
async def test_scrub():
    await FFXIAHScrubber().scrub(item_ids={1, 2, 3, 4})


@pytest.mark.asyncio
async def test_augment_item_info():
    async with aiohttp.ClientSession() as session:
        results, failed = await FFXIAHScrubber()._get_item_data(session, [4096])
        kwargs = augment_item_info(results[4096])
        assert kwargs["name"] == "Fire Crystal"
